/******************************************************************/
/*            Key protection program for em and xgeom		  */
/* Copyright (c) 1990 Sonnet Software, Inc., All rights reserved. */
/******************************************************************/

/* %W% last mod on %G% at %U% */
/* Full SCCS path name: (%P%) */

/*******************************************************************

	NAME: crypt2.i
	
	AUTHOR: K. Schroeder
	
	PURPOSE: Determine the key for the keyfile for release 2.2 and
		 beyond.
	
	METHOD:
		For release 2.2 and beyond the following is the method
		for locking and unlocking the software
		
	Pre 2.2

	Lock Software
		1. Convert key file info into a string.
		2. Make a key from string
		3. encrypt key
		4. add check sum
		5. print user key
	Unlockem	
		1. Convert key file info to a string
		2. Make a key from string
		3. decrypt user key -> internal key
		4. check checksum, if wrong print out message
		5. check interal key with key made from keyfile
	
	Release 2.2 and above	

	Lock Software
		1. Convert key file info into a string.
		2. Make a key from string
		3. encrypt using 2.2 algorithm of first x characters
		4. convert to integer
		5. Get checksum, then convert to integer
		6. print integer ey
	Unlockem
		1. Convert key file info to a string
		2. Make a key from string
		3. Get user integer key
		4. convert to alpha-numberic -> passwd1
		5. Check the checksum
		6. decrypt passwd1
		7. Compare with key.
	
	CONTENTS:

*******************************************************************/

/* definitions	*/

/* external	*/

/* structures and global/semiglobal variable */

int chr_to_value();


/*******************************************************************
	NAME: check_key2
	
	DESCRIPTION: 
		Check if the user has in the correct key.
	
	RETURNS: 
		KEY_GOOD -  user has valid key
		KEY_BAD -  user has invalid key
		KEY_WRONG_CHECSUM -  user has invalid key.  (The checksum
			on the key is incorrect.  This means the user
			either did not enter in the correct key given
			to them by Sonnet Rep.  Or they are just guessing
			at a key.)

*******************************************************************/

int
check_key2(key)
Key *key;
{
	char *string;
	char *internal_key;
	char key_str[KEY_LENGTH];
	char file_sum[3];
	char det_sum[3];
	int len;

	/*
	**	Turn key file into a string then convert to an
	**	internal key.
	*/
	write_key_string(&string,key,TRUE);
	internal_key = make_key(string);
	sFree(string);
	internal_key[7] = '\0';

	/*
	**	Decrypt the user's key
	*/
	key_to_string2(key_str,key->key);

	/*
	**	Extract check sum
	*/
	len = strlen(key_str);
	file_sum[0] = key_str[len-2];
	file_sum[1] = key_str[len-1];
	file_sum[2] = '\0';
	key_str[len-2] = '\0';

	/* check the check sum */
	get_check_sum(det_sum,key_str);

	if ( strcmp(det_sum,file_sum) != 0 ) {
		return KEY_WRONG_CHECKSUM;
	}
	if ( strncmp( key_str, internal_key, 7) == 0 ) {
		return KEY_GOOD;
	} else {
		return KEY_BAD;
	}
}

/*******************************************************************
	Convert a string 9 characters to a key (20 digits) to 
*******************************************************************/
void
string_to_key2(key,str)
char *key;
char *str;
{
	int block[18];
	int i;

	for (i=0; i < 3; ++i) {
		key_to_block2(block,str+i*3);
		encrypt2(block);
		block_to_number2(key+i*6,block);
		key[18] = '\0';
	}
}

/*******************************************************************
	Convert a key (20 digits) to a string 9 characters
*******************************************************************/
void
key_to_string2(str,key)
char *str;
char *key;
{
	int block[18];
	int i;

	for (i=0; i < 3; ++i) {
		number_to_block2(block,key+i*6);
		decrypt2(block);
		block_to_key2(str+i*3,block);
		str[9] = '\0';
	}
}

/*******************************************************************
	Convert number to block of 18 bits.
*******************************************************************/
void
block_to_number2(num,block)
char *num;
int *block;
{
	int byte;
	int num_value;
	
	for (byte=0; byte<6; ++byte) {
		num_value = block[byte*3+2] * 4 +
			    block[byte*3+1] * 2 +
			    block[byte*3+0];
		num[byte] = '2' + num_value;
	}
}

/*******************************************************************
	Convert a block of 18 bits to 6 number from 2-9.
*******************************************************************/
void
number_to_block2(block,num)
int *block;
char *num;
{
	int byte;
	int num_value;
	int bit;
	int div;
	int nbit;
	int j;
	
	for (byte=0; byte<6; ++byte) {
		num_value = num[byte] - '2';
		for (bit=0; bit < 3; ++bit ) {
			div = 1;
			for (j=0; j<bit; ++j) div *= 2;
			nbit = num_value/div - 2*(num_value/ (2*div) );
			block[byte*3+bit] = nbit;
		}
	}
}

/*******************************************************************
	Changes a key to a block containing 1's and 0's	
*******************************************************************/
void
key_to_block2(block,key)
int *block;
char *key;
{
	int byte;
	int bit;
	int char_value;
	int nbit;
	int j;
	int div;
	
	for (byte=0; byte<3; ++byte) {
		char_value = chr_to_value( key[byte] );
		for (bit=0; bit < 6; ++bit ) {
			div = 1;
			for (j=0; j<bit; ++j) div *= 2;
			nbit = char_value/div - 2*(char_value/ (2*div) );
			block[byte*6+bit] = nbit;
		}
	}
}

/*******************************************************************
	Changes a block containing 1's and 0's to a key
*******************************************************************/
void
block_to_key2(key,block)
char *key;
int *block;
{
	int byte;
	int char_value;
	int bit;
	int j;
	int mult;

	for (byte=0; byte<3; ++byte) {
		char_value = 0;
		for (bit=0; bit < 6; ++bit ) {
			mult = 1;
			for (j=0; j<bit; ++j) mult *= 2;
			char_value += mult * block[byte*6+bit];
		}
		value_to_char( key+byte, char_value );
	}
}

/*******************************************************************
	Take a character and convert it to a integer
	Mapping is:
		a-z	->	0-25
		A-Z	->	26-51
		0-9	->	52-61
		/	->	62
		.	->	63
*******************************************************************/
int
chr_to_value(c) 
char c;
{
	int value;

	if ( c >='a' && c <= 'z' ) {
		value = c - 'a' + 0;
	} else if ( c >='A' && c <= 'Z' ) {
		value = c - 'A' + 26;
	} else if ( c >='0' && c <= '9' ) {
		value = c - '0' + 52;
	} else if ( c == '/' ) {
		value = 62;
	} else if ( c == '.' ) {
		value = 63;
	} else {
		fprintf(stderr,"Improper character = %c (%o)\n",c,(int)c);
		exit (1);
	}
	
	return (value);
}

/*******************************************************************
	Convert integer in value to a character
*******************************************************************/
void
value_to_char( cpnt, value )
char *cpnt;
int value;
{
	char chr;
	
	if      ( value <= 25 ) chr = (char)('a' + value);
	else if ( value <= 51 ) chr = (char)('A' + value - 26);
	else if ( value <= 61 ) chr = (char)('0' + value - 52);
	else if ( value == 62 ) chr = '/'; /* need "./" for salt */
	else if ( value == 63 ) chr = '.';

	*cpnt = chr;
}

struct {
	int n;
	int beg;
	int len;
	int invert;
} SB[] = { 
	12,  0,  2,  0,
	12,  0,  8,  0,
	10, 11,  6,  0,
	12, 13, 14,  1,
	14,  6,  8,  0,
	 7, 16,  9,  1,
	 8,  9,  3,  0,
	 3,  6,  6,  1,
	 7, 11, 13,  1,
	14,  6, 14,  1,
	11,  8,  6,  1,
	11,  3,  7,  0,
	12, 10,  2,  1,
	 8, 12, 13,  1,
	 6,  3, 12,  1,
	 8, 10,  8,  1,
	11, 16,  2,  0,
	 6,  2,  8,  0,
	 4,  3,  6,  1,
	10,  0, 10,  0,
	14,  5, 11,  0,
	 5,  3,  2,  1,
	 6, 14,  4,  0,
	 3, 15,  5,  1,
	 8, 11, 13,  1,
	 9, 11,  6,  1,
	15, 15,  2,  0,
	15, 15, 13,  0,
	 6, 15,  4,  0,
	12, 15,  3,  1,
	 2,  0, 12,  0,
	 7, 14,  9,  1,
	10,  6, 14,  0,
	12,  2, 12,  1,
	 3, 14, 11,  1,
	12,  5,  6,  1,
	10, 16, 10,  1,
	 8, 12,  3,  0,
	13,  2,  3,  1,
	 8, 14, 10,  0,
	 4,  3,  9,  1,
	 3, 11, 10,  1,
	14, 15, 11,  1,
	12,  4,  5,  0,
	 8,  3,  6,  1,
	15,  2, 11,  1,
	15,  9, 12,  0,
	 8,  6,  9,  1,
	14,  0,  2,  1,
	 9, 15,  6,  1,
	10,  1, 11,  0,
	 3,  8, 10,  1,
	13,  3,  7,  1,
	11,  5,  5,  1,
	 9,  2, 11,  0,
	 9, 11,  4,  0,
	10, 14,  2,  1,
	 3, 13,  8,  0,
	13,  5, 14,  0,
	15,  7,  8,  0,
	12,  6, 14,  0,
	14,  0,  6,  0,
	 6,  8,  9,  1,
	 8,  9,  3,  0,
	 3, 14,  5,  1,
	10,  5, 14,  0,
	 2, 14,  2,  0,
	 6,  3, 14,  0,
	13, 11,  4,  1,
	 6, 14, 12,  1,
	 2, 14,  9,  0,
	 3,  3,  6,  1,
	13, 16,  9,  1,
	16,  4,  6,  0,
	 3,  6, 11,  0,
	 8, 16,  6,  0,
	 7,  7, 14,  1,
	11, 15, 12,  1,
	 6,  2,  2,  1,
	 4,  3,  3,  0,
	 5,  2,  7,  1,
	 3,  7,  5,  1,
	10,  1,  8,  0,
	 6, 12, 14,  0,
	 7,  5,  6,  0,
	11, 11,  2,  0,
	10,  0,  6,  0,
	 4,  9,  8,  0,
	 8, 11,  6,  0,
	14,  7,  9,  1,
	13,  9,  5,  0,
	 6,  1,  2,  1,
	 2, 11,  2,  0,
	 8, 16, 13,  1,
	12,  4,  2,  1,
	 4,  2, 13,  1,
	14,  6,  7,  0,
	 7, 15,  2,  1,
	 3,  4,  8,  1,
	 5,  5,  8,  0,
};

/*******************************************************************
	Encrypt a block
*******************************************************************/
void
encrypt2(block)
int *block;
{
	int i;
	int sb_size;
	
	sb_size = sizeof(SB)/sizeof(*SB);
	for (i=0; i<sb_size; ++i ) {
		shift_block(block,SB[i].n,SB[i].beg,SB[i].len,SB[i].invert);
	}
}

/*******************************************************************
	Decrypt a block
*******************************************************************/
void
decrypt2(block)
int *block;
{
	int i;
	int sb_size;

	sb_size = sizeof(SB)/sizeof(*SB);
	for (i=sb_size-1; i>=0; --i ) {
		shift_block(block,-SB[i].n,SB[i].beg,SB[i].len,SB[i].invert);
	}
}

/*************************************************************************
	Shift block by n-bits starting at bit beg for len number of bits.
*************************************************************************/
void
shift_block(block,n,beg, len, invert)
int *block;
int n;
int beg;
int len;
int invert;
{
	int nblock[18];
	int i,j;
	int end;
	int index_i, index_j;

	cpblock(nblock,block);

	end = beg + len;
	for ( i = beg; i < end; ++i ) {
		j = i+n;
		index_i = get_index(i,beg,len);
		index_j = get_index(j,beg,len);
		nblock[index_i] = block[index_j];
		if (invert) {
			if ( nblock[index_i] == 0 ) {
				nblock[index_i] = 1;
			} else {
				nblock[index_i] = 0;
			}
		}
	}

	cpblock(block,nblock);
}

/*******************************************************************
	return n+len, taking into wraparound.
	
	n is an index into a block of bit.  With in the block is a
	subset of bits beginning at "beg" for "len" bit, where the 
	subset of bit may wrap around the end of the block.
	This routine returns n+beg bits keeping within the subset of
	bit taking into account the wrap around.  n may be either positive
	or negative and may be larger than len.
*******************************************************************/
int
get_index(n,beg,len)
int n;
int beg;
int len;
{
	n -= beg;
	while ( n < 0 ) n += len;
	while ( n >= len ) n -= len;
	n += beg;
        if ( n >= 18 ) n -= 18;
	return (n);
}

/*******************************************************************
	copies block of data
*******************************************************************/
void
cpblock(to,from)
int *to;
int *from;
{
	int i;
	for (i=0; i<18; ++i) {
		to[i] = from[i];
	}
}

